@file:Suppress("unused", "MemberVisibilityCanBePrivate")

package no.vegvesen.meldingsformidler

import com.fasterxml.jackson.core.util.VersionUtil
import com.fasterxml.jackson.databind.DeserializationFeature
import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule
import com.fasterxml.jackson.module.kotlin.readValue
import com.fasterxml.jackson.module.kotlin.registerKotlinModule
import no.vegvesen.meldingsformidler.common.domain.Fodselsnummer
import no.vegvesen.meldingsformidler.common.domain.Kanal
import no.vegvesen.meldingsformidler.common.domain.MeldingUuid
import no.vegvesen.meldingsformidler.common.filter.ContentIdFilter.Companion.CONTENT_ID_HEADER
import no.vegvesen.meldingsformidler.common.openamclient.OpenAmClient
import no.vegvesen.meldingsformidler.send.ClientError
import no.vegvesen.meldingsformidler.send.DokumentMelding
import no.vegvesen.meldingsformidler.send.EpostMelding
import no.vegvesen.meldingsformidler.send.IdempotencyKey
import no.vegvesen.meldingsformidler.send.Melding
import no.vegvesen.meldingsformidler.send.MeldingsformidlerResult
import no.vegvesen.meldingsformidler.send.MeldingsformidlerResult.Companion.failure
import no.vegvesen.meldingsformidler.send.MeldingsformidlerResult.Companion.success
import no.vegvesen.meldingsformidler.send.SentMelding
import no.vegvesen.meldingsformidler.send.SentMelding.Companion.toInternal
import no.vegvesen.meldingsformidler.send.UnexpectedError
import no.vegvesen.meldingsformidler.send.toDto
import no.vegvesen.meldingsformidler.send.toHash
import no.vegvesen.meldingsformidler.status.StatusMelding
import no.vegvesen.meldingsformidler.status.StatusMelding.Companion.toInternal
import no.vegvesen.mottak.rest.dto.KanalForMottakerResponseV2
import no.vegvesen.mottak.rest.dto.MeldingStatusResponseV2
import no.vegvesen.mottak.rest.dto.ProblemDetail
import no.vegvesen.mottak.rest.dto.SendDokumentMeldingResponseV2
import no.vegvesen.mottak.rest.dto.SendEpostResponseV2
import java.net.ConnectException
import java.net.URI
import java.net.http.HttpClient
import java.net.http.HttpRequest
import java.net.http.HttpResponse
import java.util.Properties
import javax.ws.rs.core.HttpHeaders
import javax.ws.rs.core.MediaType


@Suppress("unused")
class MeldingsformidlerClient private constructor(
    private val baseUri: URI,
    private val openAmClient: OpenAmClient?,
    private val httpClient: HttpClient,
) {

    constructor(
        environment: MeldingsformidlerEnvironment,
        openAmClient: OpenAmClient? = null,
        httpClient: HttpClient = defaultHttpClient,
    ) : this(
        openAmClient = openAmClient,
        baseUri = environment.baseUri,
        httpClient = httpClient,
    )

    private val objectMapper = ObjectMapper()
        .registerKotlinModule()
        .registerModule(JavaTimeModule())
        .configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false)

    /**
     *
     * @param melding The message to be sent
     * @param idempotencyKey Uniquely identifies the message. If a message has been sent before with the same idempotency key, the message will not be sent again.
     */
    fun send(melding: Melding, idempotencyKey: IdempotencyKey): MeldingsformidlerResult<SentMelding> =
        when (melding) {
            is DokumentMelding -> sendDokumentMelding(melding, idempotencyKey)
            is EpostMelding -> sendEpost(melding, idempotencyKey)
        }

    private fun sendDokumentMelding(dokumentMelding: DokumentMelding, idempotencyKey: IdempotencyKey): MeldingsformidlerResult<SentMelding> {
        return runCatchingMeldingsformidlerResult {
            val request = requestBuilderBase()
                .header(IDEMPOTENCY_HEADER, idempotencyKey.value)
                .header(CLIENT_VERSION_HEADER, getClientVersion())
                .headers(CONTENT_ID_HEADER, dokumentMelding.toDto().toHash())
                .uri(baseUri.resolve("/api/v2/meldinger/send/dokumentmelding"))
                .POST(HttpRequest.BodyPublishers.ofString(objectMapper.writeValueAsString(dokumentMelding.toDto())))
                .build()

            val response: HttpResponse<String> =
                httpClient
                    .send(request, HttpResponse.BodyHandlers.ofString())

            return parseResult<SendDokumentMeldingResponseV2>(response)
                .fold(
                    onSuccess = { success(it.toInternal()) },
                    onFailure = { failure(it.error) }
                )
        }
    }

    private fun sendEpost(epostMelding: EpostMelding, idempotencyKey: IdempotencyKey): MeldingsformidlerResult<SentMelding> {
        return runCatchingMeldingsformidlerResult {
            val request = requestBuilderBase()
                .header(IDEMPOTENCY_HEADER, idempotencyKey.value)
                .header(CLIENT_VERSION_HEADER, getClientVersion())
                .headers(CONTENT_ID_HEADER, epostMelding.toDto().toHash())
                .uri(baseUri.resolve("/api/v2/meldinger/send/epost"))
                .POST(HttpRequest.BodyPublishers.ofString(objectMapper.writeValueAsString(epostMelding.toDto())))
                .build()

            val response: HttpResponse<String> =
                httpClient
                    .send(request, HttpResponse.BodyHandlers.ofString())

            return parseResult<SendEpostResponseV2>(response)
                .fold(
                    onSuccess = { success(it.toInternal()) },
                    onFailure = { failure(it.error) }
                )
        }
    }

    fun status(meldingUuid: MeldingUuid): MeldingsformidlerResult<StatusMelding> {
        return runCatchingMeldingsformidlerResult {
            val request = requestBuilderBase()
                .uri(baseUri.resolve("/api/v2/meldinger/${meldingUuid.id}/status"))
                .GET()
                .build()

            val response: HttpResponse<String> = httpClient
                .send(request, HttpResponse.BodyHandlers.ofString())

            return parseResult<MeldingStatusResponseV2>(response)
                .fold(
                    onSuccess = { success(it.toInternal()) },
                    onFailure = { failure(it.error) }
                )
        }
    }

    fun findCurrentKanalFor(personIdentifikator: Fodselsnummer): MeldingsformidlerResult<Kanal> {
        return runCatchingMeldingsformidlerResult {
            val request = requestBuilderBase()
                .uri(baseUri.resolve("/api/v2/person/${personIdentifikator.value}/kanal"))
                .GET()
                .build()
            val response: HttpResponse<String> = httpClient
                .send(request, HttpResponse.BodyHandlers.ofString())

            return parseResult<KanalForMottakerResponseV2>(response)
                .fold(
                    onSuccess = { success(it.toInternal()) },
                    onFailure = { failure(it.error) }
                )
        }
    }

    private fun KanalForMottakerResponseV2.toInternal(): Kanal =
        when (this.kanal) {
            KanalForMottakerResponseV2.Kanal.DigitalPost -> Kanal.DigitalPost
            KanalForMottakerResponseV2.Kanal.Brev -> Kanal.Brev
            KanalForMottakerResponseV2.Kanal.Altinn -> Kanal.Altinn
        }

    private inline fun <reified T> parseResult(response: HttpResponse<String>): MeldingsformidlerResult<T> = runCatching {
        success(
            objectMapper
                .readValue<T>(response.body())
        )
    }.recover {
        failure(ClientError(objectMapper.readValue<ProblemDetail>(response.body())))
    }.getOrElse {
        failure(ClientError(errorCodeAsText(response.statusCode(), it)))
    }

    private fun errorCodeAsText(statusCode: Int, exception: Throwable): ProblemDetail {
        val errorAsDescription = when (statusCode) {
            401 -> "The autorization token is not valid. Please verify that you have setup OpenAmClient or your own authorization implementation, and make sure the tjenestebruker has the right roles"
            404 -> "Did not find the endpoint for sending melding. If you did override the base Uri, please verify that it is correct. Otherwise, the system might be unavailable."
            500 -> "An internal server error occurred. Please contact us on #meldingsformilding-brukerforum."
            else -> throw exception
        }
        return ProblemDetail(
            status = statusCode,
            detail = errorAsDescription,
        )
    }

    fun health(): MeldingsformidlerEndpointHealth {
        val request = requestBuilderBase()
            .uri(baseUri.resolve("/api/v2/health"))
            .GET()
            .build()

        return try {
            httpClient
                .send(request, HttpResponse.BodyHandlers.ofString())
                .let { MeldingsformidlerEndpointHealth.fromStatusCode(it.statusCode()) }
        } catch (exception: ConnectException) {
            MeldingsformidlerEndpointHealth.NotReachable
        }
    }

    private fun getClientVersion(): String {
        val properties = Properties()
        val inputStream = VersionUtil::class.java.classLoader.getResourceAsStream("version.properties")
        properties.load(inputStream)
        return properties.getProperty("version", "unknown")
    }


    private fun requestBuilderBase(): HttpRequest.Builder {
        val builder = HttpRequest.newBuilder()
            .header(HttpHeaders.CONTENT_TYPE, MediaType.APPLICATION_JSON)
            .header(HttpHeaders.ACCEPT, MediaType.APPLICATION_JSON)

        return when (openAmClient) {
            null -> builder
            else -> builder.header(HttpHeaders.AUTHORIZATION, "Bearer ${openAmClient.token().token}")
        }
    }

    private inline fun <T> runCatchingMeldingsformidlerResult(block: () -> MeldingsformidlerResult<T>): MeldingsformidlerResult<T> =
        runCatching(block)
            .getOrElse { e ->
                return when (e) {
                    is ClientErrorException -> failure(ClientError(ProblemDetail(detail = e.description)))
                    else -> failure(UnexpectedError(e))
                }
            }


    companion object {
        val defaultHttpClient: HttpClient = HttpClient.newBuilder()
            .version(HttpClient.Version.HTTP_2)
            .build()
        const val IDEMPOTENCY_HEADER = "Idempotency-Key"
        const val CLIENT_VERSION_HEADER = "Client-Version"
    }
}
