package no.vegvesen.meldingsformidler.send

import no.vegvesen.meldingsformidler.extensions.StringExtensions.requiredLength
import no.vegvesen.meldingsformidler.common.domain.ArligRepeterendePeriode
import no.vegvesen.meldingsformidler.common.domain.ColorHex
import no.vegvesen.meldingsformidler.common.domain.Fodselsnummer
import no.vegvesen.meldingsformidler.common.domain.Gyldighetsperiode
import no.vegvesen.meldingsformidler.common.domain.Period
import no.vegvesen.meldingsformidler.common.domain.Sprak
import no.vegvesen.mottak.rest.dto.ArligRepeterendePeriodeTimeOfMonthV2Dto
import no.vegvesen.mottak.rest.dto.ArligRepeterendePeriodeValidMonthsV2Dto
import no.vegvesen.mottak.rest.dto.ArligRepeterendePeriodeValidYearsV2Dto
import no.vegvesen.mottak.rest.dto.BevisGyldighetsperiodeArligRepeterendePeriodeV2Dto
import no.vegvesen.mottak.rest.dto.BevisGyldighetsperiodePeriodeV2Dto
import no.vegvesen.mottak.rest.dto.DokumentMeldingBevisAdresseV2Dto
import no.vegvesen.mottak.rest.dto.DokumentMeldingBevisAttributtV2Dto
import no.vegvesen.mottak.rest.dto.DokumentMeldingBevisGyldighetsperiodeV2Dto
import no.vegvesen.mottak.rest.dto.DokumentMeldingBevisHolderV2Dto
import no.vegvesen.mottak.rest.dto.DokumentMeldingBevisInfoV2Dto
import no.vegvesen.mottak.rest.dto.DokumentMeldingBevisV2Dto
import java.time.Instant

data class Bevis(
    val bevisTittel: String,
    val bevisHolder: BevisHolder,
    val sprak: Sprak,
    val utstedtTidspunkt: Instant? = null,
    val gyldighetsperiode: Gyldighetsperiode? = null,
    val attributtList: List<BevisAttributt> = emptyList(),
    val infoList: List<BevisInfo> = emptyList(),
    val utstederVisningsnavn: String = "Statens vegvesen",
    val bakgrunnsfarge: ColorHex = ColorHex("#FFFFFF"),
) {
    init {
        validateAllFields(
            utstederVisningsnavn = utstederVisningsnavn,
            bevisTittel = bevisTittel,
            attributtListSize = attributtList.size,
            infoListSize = infoList.size
        )
    }

    companion object {
        const val maxUtstederVisningsnavnLength = 100
        const val maxBevisTittelLength = 255
        const val maxAttributtSize = 5
        const val maxInfoSize = 3

        fun validateAllFields(utstederVisningsnavn: String, bevisTittel: String, attributtListSize: Int, infoListSize: Int) {
            utstederVisningsnavn.requiredLength(1..maxUtstederVisningsnavnLength, "UtstederVisningsnavn")
            bevisTittel.requiredLength(1..maxBevisTittelLength, "Tittel")
            attributtListSize.requiredLength(0..maxAttributtSize, "Attributt list")
            infoListSize.requiredLength(0..maxInfoSize, "Info list")
        }
    }

}

data class BevisHolder(
    val name: Name,
    val fodselsnummer: Fodselsnummer,
    val bevisAdresse: BevisAdresse?
)

data class Name(
    val first: String,
    val last: String
) {
    init {
        validateAllNameFields(
            first = first,
            last = last
        )
    }

    companion object {
        const val maxNameLength = 500

        fun validateAllNameFields(first: String, last: String) {
            first.requiredLength(1..maxNameLength, "First name")
            last.requiredLength(1..maxNameLength, "Last name")
        }
    }
}

data class BevisAdresse(
    val gate: String,
    val postnummer: String,
    val poststed: String,
    val land: String
)

data class BevisAttributt(
    val name: String,
    val value: String
) {
    init {
        validateAllAttributtFields(
            name = name,
            value = value
        )
    }

    companion object {
        const val maxAttributtNameLength = 100
        const val maxAttributtValueLength = 100

        fun validateAllAttributtFields(name: String, value: String) {
            name.requiredLength(1..maxAttributtNameLength, "Attributt name")
            value.requiredLength(1..maxAttributtValueLength, "Attributt value")
        }
    }

}

data class BevisInfo(
    val name: String,
    val value: String
) {
    init {
        validateAllInfoFields(
            name = name,
            value = value
        )
    }

    companion object {
        const val maxInfoNameLength = 100
        const val maxInfoValueLength = 255

        fun validateAllInfoFields(name: String, value: String) {
            name.requiredLength(1..maxInfoNameLength, "Info name")
            value.requiredLength(1..maxInfoValueLength, "Info value")
        }
    }
}

internal fun Bevis.toDto(): DokumentMeldingBevisV2Dto =
    DokumentMeldingBevisV2Dto(
        utstedttidspunkt = utstedtTidspunkt,
        bevisHolder = DokumentMeldingBevisHolderV2Dto(
            fornavn = bevisHolder.name.first,
            etternavn = bevisHolder.name.last,
            fodselsnummer = bevisHolder.fodselsnummer.value,
            adresse = bevisHolder.bevisAdresse?.let {
                DokumentMeldingBevisAdresseV2Dto(
                    gate = it.gate,
                    postnummer = it.postnummer,
                    poststed = it.poststed,
                    land = it.land
                )
            }
        ),
        bevisTittel = bevisTittel,
        attributtList = attributtList.map {
            DokumentMeldingBevisAttributtV2Dto(
                bevisAttributtNavn = it.name,
                bevisAttributtVerdi = it.value
            )
        },
        sprak = when (sprak) {
            Sprak.Bokmal -> DokumentMeldingBevisV2Dto.Sprak.Bokmal
            Sprak.Nynorsk -> DokumentMeldingBevisV2Dto.Sprak.Nynorsk
            Sprak.Engelsk -> DokumentMeldingBevisV2Dto.Sprak.Engelsk
            Sprak.Samisk -> DokumentMeldingBevisV2Dto.Sprak.Samisk
        },
        infoList = infoList.map {
            DokumentMeldingBevisInfoV2Dto(
                bevisInfoNavn = it.name,
                bevisInfoVerdi = it.value
            )
        },
        utstederVisningsnavn = utstederVisningsnavn,
        bakgrunnsfarge = bakgrunnsfarge.value,
        gyldighetsperiode = gyldighetsperiode?.let {
            when (it) {
                is ArligRepeterendePeriode ->
                    DokumentMeldingBevisGyldighetsperiodeV2Dto(
                        arligRepeterendePeriode = BevisGyldighetsperiodeArligRepeterendePeriodeV2Dto(
                            validMonths = ArligRepeterendePeriodeValidMonthsV2Dto(
                                start = ArligRepeterendePeriodeTimeOfMonthV2Dto(
                                    time = it.validMonths.start.time.toString(),
                                    monthDay = it.validMonths.start.monthDay.toString()
                                ),
                                endInclusive = ArligRepeterendePeriodeTimeOfMonthV2Dto(
                                    time = it.validMonths.endInclusive.time.toString(),
                                    monthDay = it.validMonths.endInclusive.monthDay.toString()
                                )
                            ),
                            validYears = ArligRepeterendePeriodeValidYearsV2Dto(
                                start = it.validYears?.start?.value.toString(),
                                endInclusive = it.validYears?.endInclusive?.value.toString()
                            )
                        )
                    )

                is Period -> DokumentMeldingBevisGyldighetsperiodeV2Dto(
                    periode = BevisGyldighetsperiodePeriodeV2Dto(
                        start = it.validRange.start,
                        endInclusive = it.validRange.endInclusive
                    )
                )
            }
        }
    )