package no.vegvesen.meldingsformidler.send

import no.vegvesen.meldingsformidler.extensions.StringExtensions.requiredLength
import no.vegvesen.meldingsformidler.common.domain.AdresseLinjer
import no.vegvesen.meldingsformidler.common.domain.AdresseType
import no.vegvesen.meldingsformidler.common.domain.AltinnTjenestedefinisjon
import no.vegvesen.meldingsformidler.common.domain.DigitalPostVarsel
import no.vegvesen.meldingsformidler.common.domain.IktLosningId
import no.vegvesen.meldingsformidler.common.domain.IktLosningReferanse
import no.vegvesen.meldingsformidler.common.domain.Postadresse
import no.vegvesen.meldingsformidler.common.domain.PostadresseNorsk
import no.vegvesen.meldingsformidler.common.domain.Returhandtering
import no.vegvesen.meldingsformidler.common.domain.Sikkerhetsniva
import no.vegvesen.meldingsformidler.common.domain.Sprak
import no.vegvesen.meldingsformidler.common.domain.Tjenestekode
import no.vegvesen.meldingsformidler.common.domain.Tjenesteutgave
import no.vegvesen.mottak.rest.dto.AvsenderAdresseLinjerV2Dto
import no.vegvesen.mottak.rest.dto.DokumentMeldingAltinnConfigurationV2Dto
import no.vegvesen.mottak.rest.dto.DokumentMeldingAvsenderAdresseV2Dto
import no.vegvesen.mottak.rest.dto.DokumentMeldingConfigurationV2Dto
import no.vegvesen.mottak.rest.dto.DokumentMeldingOffentligPostConfigurationV2Dto
import java.time.LocalDate

sealed interface Configuration {
    val iktLosningId: IktLosningId
}

data class DokumentConfiguration(
    override val iktLosningId: IktLosningId,
    val sprak: Sprak,
    val altinnConfiguration: AltinnConfiguration = defaultAltinnConfiguration,
    val offentligPostConfiguration: OffentligPostConfiguration = defaultOffentligPostConfiguration,
) : Configuration {

    companion object {
        val defaultOffentligPostConfiguration = OffentligPostConfiguration()
        val defaultAltinnConfiguration = AltinnConfiguration()
    }
}

data class EpostConfiguration(override val iktLosningId: IktLosningId) : Configuration

/**
 * Configuration for public post.
 *
 * @property sikkerhetsniva The security level. Defaults to `Hoyeste`.
 * @property virkningsdato The effective date. Defaults to `null` and must be within 5 days in the future.
 * @property returhandtering The return handling method. Defaults to `MakuleresAvPrintLeverandor`.
 * @property iktlosningReferanse The `iktlosningReferanse` can be a case number or similar that can be used to trace the origin of the message in the consuming system.
 * @property avsender is the one who will be listed as the sender of the given message.
 * @property digitalPostVarsel Policy for how the receiver will be notified
 */

data class OffentligPostConfiguration(
    val sikkerhetsniva: Sikkerhetsniva = defaultSikkerhetsniva,
    val virkningsdato: LocalDate? = defaultVirkningsdato,
    val returhandtering: Returhandtering = defaultReturhandtering,
    val iktlosningReferanse: IktLosningReferanse? = defaultIktLosningsReferanse,
    val avsender: Postadresse = defaultAvsender,
    val digitalPostVarsel: DigitalPostVarsel = defaultDigitalPostVarsel
) {
    init {
        validateAllOffentligPostConfigurationFields(iktlosningReferanse, virkningsdato)
    }

    companion object {
        val defaultSikkerhetsniva: Sikkerhetsniva = Sikkerhetsniva.Hoyeste
        val defaultVirkningsdato: LocalDate? = null
        val defaultIktLosningsReferanse: IktLosningReferanse? = null
        val defaultReturhandtering: Returhandtering = Returhandtering.Makulere
        val defaultDigitalPostVarsel: DigitalPostVarsel = DigitalPostVarsel.VarselFraPostkasse
        val defaultAvsender: Postadresse = PostadresseNorsk(
            navn = "Statens vegvesen",
            adresseLinjer = AdresseLinjer.from(listOf("Postboks 1010 Nordre Ål")),
            adresseType = AdresseType.Avsender,
            postNummer = "2605",
            postSted = "Lillehammer",
        )

        const val MAX_IKT_LOSNING_REFERANSE_LENGTH = 40
        const val VIRKNINGSDATO_WITHIN: Long = 5

        fun validateAllOffentligPostConfigurationFields(iktLosningReferanse: IktLosningReferanse?, virkningsdato: LocalDate?) {
            require(iktLosningReferanse == null || iktLosningReferanse.value.length <= MAX_IKT_LOSNING_REFERANSE_LENGTH) { "IktlosningReferanse can be maximum $MAX_IKT_LOSNING_REFERANSE_LENGTH characters" }
            require(
                virkningsdato == null || virkningsdato.isBefore(
                    LocalDate.now().plusDays(VIRKNINGSDATO_WITHIN)
                )
            ) { "Virkningsdato must be max $VIRKNINGSDATO_WITHIN days from today." }
        }
    }
}

/**
 * Configuration for Altinn.
 *
 * @property tjenestedefinisjon Default value is 2762/201109 (Brev fra Statens vegvesen). To set up a new service: https://altinn.github.io/docs/tul/tjenestetyper/ny/
 */
data class AltinnConfiguration(
    val tjenestedefinisjon: AltinnTjenestedefinisjon = defaultTjenestedefinisjon,
    val tillatVideresending: Boolean = defaultTillatVideresending,
    val beskrivelse: String = defaultBeskrivelse,
) {
    init {
        validateAllAltinnConfigurationFields(beskrivelse)
    }

    companion object {
        val defaultTjenestedefinisjon = AltinnTjenestedefinisjon(
            tjenestekode = Tjenestekode("2762"),
            tjenesteutgave = Tjenesteutgave("201109")
        )
        const val defaultTillatVideresending = false
        const val defaultBeskrivelse = ""

        const val maxBeskrivelseLength = 512

        fun validateAllAltinnConfigurationFields(beskrivelse: String) {
            beskrivelse.requiredLength(0..maxBeskrivelseLength, "Beskrivelse")
        }
    }
}

internal fun DokumentConfiguration.toDto(): DokumentMeldingConfigurationV2Dto =
    DokumentMeldingConfigurationV2Dto(
        iktLosning = iktLosningId.value,
        sprak = when (sprak) {
            Sprak.Bokmal -> DokumentMeldingConfigurationV2Dto.Sprak.Bokmal
            Sprak.Nynorsk -> DokumentMeldingConfigurationV2Dto.Sprak.Nynorsk
            Sprak.Engelsk -> DokumentMeldingConfigurationV2Dto.Sprak.Engelsk
            Sprak.Samisk -> DokumentMeldingConfigurationV2Dto.Sprak.Samisk
        },
        altinnConfiguration = altinnConfiguration.toDto(),
        offentligPostConfiguration = offentligPostConfiguration.toDto()
    )

internal fun AltinnConfiguration.toDto(): DokumentMeldingAltinnConfigurationV2Dto =
    DokumentMeldingAltinnConfigurationV2Dto(
        beskrivelse = beskrivelse,
        tjenestekode = tjenestedefinisjon.tjenestekode.value,
        tjenesteutgave = tjenestedefinisjon.tjenesteutgave.value,
        tillatVideresending = tillatVideresending,
    )

internal fun OffentligPostConfiguration.toDto(): DokumentMeldingOffentligPostConfigurationV2Dto =
    DokumentMeldingOffentligPostConfigurationV2Dto(
        avsender = DokumentMeldingAvsenderAdresseV2Dto(
            adresseLinjer = AvsenderAdresseLinjerV2Dto(
                linje1 = avsender.adresseLinjer.linje1,
                linje2 = avsender.adresseLinjer.linje2,
                linje3 = avsender.adresseLinjer.linje3,
                linje4 = avsender.adresseLinjer.linje4,
            ),
            navn = avsender.navn,
            postNummer = (avsender as PostadresseNorsk).postNummer,
            postSted = avsender.postSted
        ),
        sikkerhetsniva = when (sikkerhetsniva) {
            Sikkerhetsniva.Betydelig -> DokumentMeldingOffentligPostConfigurationV2Dto.Sikkerhetsniva.Betydelig
            Sikkerhetsniva.Hoyeste -> DokumentMeldingOffentligPostConfigurationV2Dto.Sikkerhetsniva.Hoyeste
        },
        virkningsdato = virkningsdato,
        returhandtering = when (returhandtering) {
            Returhandtering.Returnere -> DokumentMeldingOffentligPostConfigurationV2Dto.Returhandtering.Returnere
            Returhandtering.Makulere -> DokumentMeldingOffentligPostConfigurationV2Dto.Returhandtering.Makulere
        },
        iktlosningReferanse = iktlosningReferanse?.value,
        digitalPostVarsel = when (
            digitalPostVarsel
        ) {
            DigitalPostVarsel.VarselFraPostkasse -> DokumentMeldingOffentligPostConfigurationV2Dto.DigitalPostVarsel.VarselFraPostkasse
            DigitalPostVarsel.VarselFraPostkasseOgVegvesenet -> DokumentMeldingOffentligPostConfigurationV2Dto.DigitalPostVarsel.VarselFraPostkasseOgVegvesenet
        }
    )
