package no.vegvesen.meldingsformidler.send

import no.vegvesen.meldingsformidler.extensions.StringExtensions.requiredLength
import no.vegvesen.meldingsformidler.common.domain.Dokument
import no.vegvesen.mottak.rest.dto.DokumentMeldingDokumentV2Dto
import no.vegvesen.mottak.rest.dto.DokumentMeldingInnholdV2Dto


sealed interface Innhold

data class DokumentInnhold(
    val tittel: String,
    val hoveddokument: Dokument,
    val vedlegg: List<Dokument> = emptyList(),
    val bevis: Bevis? = null
) : Innhold {

    init {
        validateAllFields(tittel, vedlegg.size, totalDokumenterByteSize())
    }

    private fun totalDokumenterByteSize(): Int = hoveddokument.bytes.size + vedlegg.sumOf { it.bytes.size }

    companion object {
        const val maxTitleLength = 100
        const val maxVedleggCount = 200
        const val maxBytesTotalInAllDocuments = 31457280

        fun validateAllFields(tittel: String, vedleggCount: Int, allDocumentsBytesCount: Int) {
            tittel.requiredLength(1..maxTitleLength, "Title")
            vedleggCount.requiredLength(0..maxVedleggCount, "Vedlegg count")
            allDocumentsBytesCount.requiredLength(1..maxBytesTotalInAllDocuments, "Total size of documents in bytes")
        }
    }
}

data class EpostInnhold(
    val emne: String,
    val innhold: String,
) : Innhold {
    init {
        validateAllEpostFields(
            emne = emne,
            innhold = innhold
        )
    }

    companion object {
        const val maxEmneLength = 100
        const val maxInnholdLength = 60000

        fun validateAllEpostFields(emne: String, innhold: String) {
            emne.requiredLength(1..maxEmneLength, "Emne")
            innhold.requiredLength(1..maxInnholdLength, "Innhold")
        }
    }
}

internal fun DokumentInnhold.toDto(): DokumentMeldingInnholdV2Dto =
    DokumentMeldingInnholdV2Dto(
        tittel = tittel,
        hoveddokument = hoveddokument.toDto(),
        vedlegg = vedlegg.map(Dokument::toDto),
        bevis = bevis?.toDto()
    )

internal fun Dokument.toDto(): DokumentMeldingDokumentV2Dto =
    DokumentMeldingDokumentV2Dto(
        tittel = tittel,
        bytes = bytes
    )