@file:Suppress("unused")

package no.vegvesen.meldingsformidler.send

import no.vegvesen.mottak.rest.dto.ProblemDetail

sealed interface MeldingsformidlerError
data class UnexpectedError(val throwable: Throwable) : MeldingsformidlerError
data class ClientError(val problemDetail: ProblemDetail) : MeldingsformidlerError

@Suppress("unchecked_cast", "unused", "memberVisibilityCanBePrivate")
@JvmInline
value class MeldingsformidlerResult<T>(val value: Any?) {

    data class MeldingsformidlerFailure(val error: MeldingsformidlerError) {
        inline fun <T> fold(
            onClientError: (clientError: ClientError) -> T,
            onUnexpectedError: (unexpectedError: UnexpectedError) -> T
        ): T {
            return when (error) {
                is UnexpectedError -> onUnexpectedError(error)
                is ClientError -> onClientError(error)
            }
        }
    }

    fun isSuccess() = value !is MeldingsformidlerFailure
    fun isFailure() = value is MeldingsformidlerFailure

    fun getOrNull(): T? = if (isSuccess()) value as T else null
    fun failureOrNull(): MeldingsformidlerFailure? = if (isFailure()) value as MeldingsformidlerFailure else null

    fun onSuccess(block: (T) -> Unit): MeldingsformidlerResult<T> {
        if (isSuccess()) block(value as T)
        return this
    }

    fun onFailure(block: (MeldingsformidlerFailure) -> Unit): MeldingsformidlerResult<T> {
        if (isFailure()) block(value as MeldingsformidlerFailure)
        return this
    }

    fun onClientError(block: (ClientError) -> Unit): MeldingsformidlerResult<T> {
        failureOrNull()?.fold(onClientError = block, onUnexpectedError = {})
        return this
    }

    fun onUnexpectedError(block: (UnexpectedError) -> Unit): MeldingsformidlerResult<T> {
        failureOrNull()?.fold(onClientError = {}, onUnexpectedError = block)
        return this
    }

    inline fun <R> fold(
        onSuccess: (success: T) -> R,
        onFailure: (failure: MeldingsformidlerFailure) -> R
    ): R {
        return when (val error = failureOrNull()) {
            null -> onSuccess(value as T)
            else -> onFailure(error)
        }
    }

    companion object {
        fun <T> success(value: T): MeldingsformidlerResult<T> = MeldingsformidlerResult(value)
        fun <T> failure(error: MeldingsformidlerError): MeldingsformidlerResult<T> = MeldingsformidlerResult(MeldingsformidlerFailure(error))

    }
}
