package no.vegvesen.meldingsformidler.integrationtests

import io.github.cdimascio.dotenv.dotenv
import no.vegvesen.meldingsformidler.MeldingsformidlerClient
import no.vegvesen.meldingsformidler.MeldingsformidlerEndpointHealth
import no.vegvesen.meldingsformidler.MeldingsformidlerEnvironment
import no.vegvesen.meldingsformidler.common.openamclient.Oauth2Client
import no.vegvesen.meldingsformidler.common.openamclient.OpenAmClient
import no.vegvesen.meldingsformidler.send.IdempotencyKey
import no.vegvesen.meldingsformidler.status.Status
import org.assertj.core.api.Assertions.assertThat
import org.assertj.core.api.Assertions.fail
import org.junit.jupiter.api.BeforeAll
import org.junit.jupiter.api.Disabled
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.TestInstance
import org.junit.jupiter.api.parallel.Execution
import org.junit.jupiter.api.parallel.ExecutionMode
import org.junit.jupiter.params.ParameterizedTest
import org.junit.jupiter.params.provider.MethodSource

@Disabled("Requires Mottak and Meldingslogg to be running in chosen environment")
@TestInstance(TestInstance.Lifecycle.PER_CLASS)
class MeldingsformidlerClientIntegrationTest {
    private lateinit var client: MeldingsformidlerClient

    @BeforeAll
    fun setupEnvironment() {
        val environment = System.getenv("ENVIRONMENT") ?: "systemtest"
        when (environment) {
            "akseptansetest" -> {
                client = MeldingsformidlerClient(MeldingsformidlerEnvironment.Akseptansetest, openAmClient(OpenAmEnvironment.Akseptansetest))
            }

            "systemtest" -> {
                client = MeldingsformidlerClient(MeldingsformidlerEnvironment.Systemtest, openAmClient(OpenAmEnvironment.Systemtest))
            }
        }
    }

    @Test
    fun `health endpoint returns healthy`() {
        val expected = MeldingsformidlerEndpointHealth.Healthy

        val actual = client.health()

        assertThat(actual).isEqualTo(expected)
    }

    @ParameterizedTest
    @MethodSource(DIGITAL_POST)
    @Execution(ExecutionMode.CONCURRENT)
    fun `send melding with same idempotencykey returns the same melding uuid`(testMeldingInput: TestMeldingInput) {
        val idempotencyKey = IdempotencyKey("123")
        val expected = client
            .send(testMeldingInput.melding, idempotencyKey)
            .fold(onSuccess = { it.id }, onFailure = { it.error })

        client
            .send(testMeldingInput.melding, idempotencyKey)
            .fold(
                onSuccess = { assertThat(it.id).isEqualTo(expected) },
                onFailure = { failure ->
                    failure.fold(
                        onClientError = { fail(it.problemDetail.toString()) },
                        onUnexpectedError = { throw it.throwable }
                    )
                }
            )
    }

    @ParameterizedTest
    @MethodSource(DIGITAL_POST, ALTINN, BREV, EPOST)
    @Execution(ExecutionMode.CONCURRENT)
    fun `polling status on sent melding returns processing or sent`(testInputMelding: TestMeldingInput) {
        val expected = listOf(Status.Processing, Status.ProcessedAndSent)

        val response = client.send(testInputMelding.melding, IntegrationTestData.randomIdempotencyKey())
        response.fold(
            onSuccess = { sentMelding ->
                val status = client.status(sentMelding.id)
                status.fold(
                    onSuccess = { statusMelding ->
                        assertThat(statusMelding.lastHendelse.status).isIn(expected)
                        assertThat(statusMelding.lastHendelse.kanal).isEqualTo(testInputMelding.expectedKanal) },
                    onFailure = { failure ->
                        failure.fold(
                            onClientError = { fail(it.problemDetail.toString()) },
                            onUnexpectedError = { throw it.throwable }
                        )
                    }
                )
            },
            onFailure = { failure ->
                failure.fold(
                    onClientError = { fail(it.problemDetail.toString()) },
                    onUnexpectedError = { throw it.throwable }
                )
            }
        )
    }

    @ParameterizedTest
    @MethodSource(FINDKANAL)
    fun `checking curret kanal for mottaker returns MeldingsformidlerResult`(testKanalInput: TestKanalInput) {
        val response = client.findCurrentKanalFor(testKanalInput.person.personIdentifikator)
        response.fold(
            onSuccess = { kanal ->
                assertThat(kanal).isEqualTo(testKanalInput.expected)
            },
            onFailure = { failure ->
                failure.fold(
                    onClientError = { fail(it.problemDetail.toString()) },
                    onUnexpectedError = { throw it.throwable }
                )
            }
        )
    }


    companion object {
        const val DIGITAL_POST = "no.vegvesen.meldingsformidler.integrationtests.IntegrationTestInput#digitalPost"
        const val ALTINN = "no.vegvesen.meldingsformidler.integrationtests.IntegrationTestInput#altinn"
        const val BREV = "no.vegvesen.meldingsformidler.integrationtests.IntegrationTestInput#brev"
        const val EPOST = "no.vegvesen.meldingsformidler.integrationtests.IntegrationTestInput#epost"
        const val FINDKANAL = "no.vegvesen.meldingsformidler.integrationtests.IntegrationTestInput#findKanal"

        @JvmStatic
        fun openAmClient(openAmEnvironment: OpenAmEnvironment): OpenAmClient =
            dotenv { directory = openAmEnvironment.environmentPath }
                .let { environmentSecrets ->
                    OpenAmClient(
                        Oauth2Client.create(
                            openAmUrl = environmentSecrets["openAmUrl"],
                            openAmUsername = environmentSecrets["tjenestebrukerUsername"],
                            openAmPassord = environmentSecrets["tjenestebrukerPassword"],
                            oidcClientId = environmentSecrets["openAmUsername"],
                            oidcClientSecret = environmentSecrets["openAmPassword"]
                        )
                    )
                }
    }
}

@Suppress("unused")
enum class OpenAmEnvironment(val environmentPath: String) {
    Systemtest("./environment/systemtest"),
    Akseptansetest("./environment/akseptansetest")
}
