@file:Suppress("ControlFlowWithEmptyBody")

package no.vegvesen.meldingsformidler.readme.examples

// <example:fullExample>
import no.vegvesen.meldingsformidler.MeldingsformidlerClient
import no.vegvesen.meldingsformidler.MeldingsformidlerEnvironment
import no.vegvesen.meldingsformidler.common.domain.Dokument
import no.vegvesen.meldingsformidler.common.domain.IktLosningId
import no.vegvesen.meldingsformidler.common.domain.MeldingUuid
import no.vegvesen.meldingsformidler.common.domain.Sprak
import no.vegvesen.meldingsformidler.common.openamclient.Oauth2Client
import no.vegvesen.meldingsformidler.common.openamclient.OpenAmClient
import no.vegvesen.meldingsformidler.send.AltinnConfiguration
import no.vegvesen.meldingsformidler.send.ClientError
import no.vegvesen.meldingsformidler.send.DokumentConfiguration
import no.vegvesen.meldingsformidler.send.DokumentInnhold
import no.vegvesen.meldingsformidler.send.DokumentMelding
import no.vegvesen.meldingsformidler.send.EpostConfiguration
import no.vegvesen.meldingsformidler.send.EpostInnhold
import no.vegvesen.meldingsformidler.send.EpostMelding
import no.vegvesen.meldingsformidler.send.EpostMottaker
import no.vegvesen.meldingsformidler.send.IdempotencyKey
import no.vegvesen.meldingsformidler.send.Melding
import no.vegvesen.meldingsformidler.send.MeldingsformidlerResult
import no.vegvesen.meldingsformidler.send.OffentligPostConfiguration
import no.vegvesen.meldingsformidler.send.Organisasjon
import no.vegvesen.meldingsformidler.send.Person
import no.vegvesen.meldingsformidler.send.SentMelding
import no.vegvesen.meldingsformidler.send.UnexpectedError
import no.vegvesen.meldingsformidler.status.NextSenderAction
import no.vegvesen.meldingsformidler.status.Status

// <example:openAmClient>
val openAmUrl = "URL til OpenAM i miljøet du skal sende meldinger til"

val openIdClientId = " Bruk din tjeneste sitt Oauth-brukernavn"
val openIdClientSecret = "Bruk din tjeneste sitt Oauth-passord"

val tjenestebruker = "Tjenestebuker for å hente token, f.eks. TjeDinTjeneste_Melding. Denne må lages av oss."
val tjenestebrukerPassword = "Passord for tjenestebrukeren"

val openAmClient = OpenAmClient(
    Oauth2Client.create(
        openAmUrl = openAmUrl,
        openAmUsername = tjenestebruker,
        openAmPassord = tjenestebrukerPassword,
        oidcClientId = openIdClientId,
        oidcClientSecret = openIdClientSecret
    )
)
// </example:openAmClient>

// <example:clientSetup>

val meldingsformidlerClient = MeldingsformidlerClient(MeldingsformidlerEnvironment.Systemtest, openAmClient)

// </example:clientSetup>

// <example:innhold>
val dokumentInnhold: DokumentInnhold = DokumentInnhold(
    tittel = "Vedtak om veivalg for E6 Kvænangsfjellet",
    hoveddokument = Dokument(tittel = "Primært beslutningsdokument", bytes = byteArrayOf(1, 2, 3)),
    vedlegg = listOf(
        Dokument(tittel = "Referat", bytes = byteArrayOf(4, 5, 6))
    )
)
// </example:innhold>

// <example:epostInnhold>
val epostInnhold: EpostInnhold = EpostInnhold(
    emne = "En epost fra Vegvesenet",
    innhold = "Melding til deg fra Vegvesenet"
)
// </example:epostInnhold>

// <example:mottaker>
val organisasjonMottaker: Organisasjon = Organisasjon("12345678900")
val personMottaker: Person = Person("12345678900")
// </example:mottaker>

// <example:epostMottaker>
val epostMottaker: EpostMottaker = EpostMottaker("hei@example.com")
// </example:epostMottaker>

// <example:configuration>
val altinnConfiguration = AltinnConfiguration()
val offentligPostConfiguration = OffentligPostConfiguration()

val dokumentConfiguration: DokumentConfiguration = DokumentConfiguration(
    sprak = Sprak.Bokmal,
    iktLosningId = IktLosningId("<din-iktløsning-id>"),
    altinnConfiguration = altinnConfiguration,
    offentligPostConfiguration = offentligPostConfiguration
)
// </example:configuration>

// <example:epostConfiguration>
val epostConfiguration = EpostConfiguration(
    iktLosningId = IktLosningId("<din-iktløsning-id>"),
)
// </example:epostConfiguration>

// <example:buildMelding>
val dokumentMelding: DokumentMelding = DokumentMelding(dokumentInnhold, personMottaker, dokumentConfiguration)
val epostMelding: EpostMelding = EpostMelding(epostInnhold, epostMottaker, epostConfiguration)
// </example:buildMelding>

// <example:sendMelding>
fun send(melding: Melding, idempotencyKey: IdempotencyKey) {
    val meldingResult: MeldingsformidlerResult<SentMelding> = meldingsformidlerClient.send(melding, idempotencyKey)
    return meldingResult.fold(
        onSuccess = { sentMelding ->
            // Meldingen er sendt og du har fått en SentMelding som har en MeldingUuid som du kan bruke til å hente status på meldingen.
            sentMelding
        },
        onFailure = { meldingsformidlerError ->
            // Noe gikk galt i sendingen, og du har fått enten en ClientError eller en UnexpectedError som må håndteres.
            meldingsformidlerError.fold(
                onClientError = ::handleClientError,
                onUnexpectedError = ::handleUnexpectedError
            )
        }
    )
}
// </example:sendMelding>

// <example:status>
fun updatedStatus(meldingUuid: MeldingUuid) =
    meldingsformidlerClient
        .status(meldingUuid)
        .fold(
            onSuccess = { statusMelding ->
                // Statusen er hentet og du kan prosessere den videre.
                processStatus(statusMelding.lastHendelse.status)
            },
            onFailure = { meldingsformidlerError ->
                // Noe gikk galt i henting av status, og du har fått enten en ClientError eller en UnexpectedError som må håndteres.
                meldingsformidlerError.fold(
                    onClientError = ::handleClientError,
                    onUnexpectedError = ::handleUnexpectedError
                )
            }
        )
// </example:status>

// <example:handleErrors>
fun handleClientError(clientError: ClientError) {
    println(clientError.problemDetail)
    // Noe gikk galt under sending av forespørselen, men det er håndtert av meldingsformidler
    // og du får en beskrivelse av hva som gikk galt.
}

fun handleUnexpectedError(unexpectedError: UnexpectedError) {
    println(unexpectedError.throwable)
    // En uforventet feil har oppstått. Dette er en feil som ikke er håndtert av meldingsformidler,
    // og du vil få en Throwable som du kan bruke til å finne ut hva som gikk galt.
}
// </example:handleErrors>

// <example:processStatus>
fun processStatus(status: Status) {
    when (status) {
        Status.Processing -> {
            // Meldingen er under behandling og avsender trenger ikke gjøre noe.
            // Alt er som det skal. Hent ny status om en stund.
        }

        Status.TemporaryErrorRetrying -> {
            // Det har oppstått en midlertidig feil som Meldingsformidler vil prøve å rette opp selv.
            // Du som avsender trenger ikke gjøre noe foreløpig.
        }

        Status.ProcessedAndSent -> {
            // Meldingen er sendt til mottaker og meldingen er ferdig behandlet.
            // Prosessen er ferdig og du trenger ikke gjøre noe mer.
        }

        Status.UnableToSend -> {
            if (status.nextSenderAction == NextSenderAction.HandleManually) {
                // Noe uventet har skjedd som ikke kan håndteres automatisk. Ta kontakt med oss i #meldingsformidling-brukerforum
            } else if (status.nextSenderAction == NextSenderAction.PollForNextStatus) {
                // Noe uventet har skjedd og Meldingsformidler vil prøve å sende meldingen igjen.
                // Hent ny status om en stund.
            }
        }
    }
}
// </example:processStatus>
// </example:fullExample>
